#:include 'common.fypp'
module elpa_wrapper
  use constants, only: sp, dp
  use debug, only: error_t
  use elpa
  use ftimings_zsvd, only: timer_t
  use matrix
#:if defined('WITH_ELPA_OPENMP')
  use omp_lib
#:endif
  implicit none

  private
#:for id in IDS
  public :: call_elpa_${id}$
#:endfor

  contains

#:for id, dtype, prec in TYPES
    !> Computes the singular vectors and singular values of A given the polar decomposition Q.
    !>
    !> A = U * M * V^T   where M is the matrix of singular values and U = Q * V
    !>
    !> See Li et al. (2019), Algorithm 3, Step 6
    subroutine call_elpa_${id}$(matrix_a, matrix_q, singular_values, matrix_right, compute_vectors, solver, mpi_comm_parent, err,&
          & timer, debug, rank0)
      type(matrix_${id}$_t), intent(in) :: matrix_a, matrix_q
      real(${prec}$), intent(out) :: singular_values(:)
      type(matrix_${id}$_t), intent(inout) :: matrix_right  ! right singular vectors V
      logical, intent(in) :: compute_vectors
      integer, intent(in) :: solver, mpi_comm_parent
      type(error_t), intent(inout) :: err
      type(timer_t), intent(inout) :: timer
      logical, intent(in) :: debug, rank0

      integer :: success
      type(matrix_${id}$_t) :: matrix_h
      class(elpa_t), pointer :: e

      call matrix_h%alloc_like(matrix_a, err, matrix_a%cols, matrix_a%cols)
      if (err%has_error()) then
        call err%add_error("Error during allocation of matrix H")
        call timer%stop()
        return
      end if

      call matrix_h%transp_mm(matrix_q, matrix_a) ! Q^T * A

      if (elpa_init(20171201) /= ELPA_OK) then
        call err%add_error("ELPA API version not supported")
        call timer%stop()
        return
      end if

      e => elpa_allocate(success)
      if (has_error(success, err, timer)) return

      call e%set("na", matrix_h%rows, success)
      if (has_error(success, err, timer)) return

      call e%set("nev", min(size(singular_values), matrix_h%rows), success)
      if (has_error(success, err, timer)) return

      call e%set("local_nrows", matrix_h%na_rows, success)
      if (has_error(success, err, timer)) return

      call e%set("local_ncols", matrix_h%na_cols, success)
      if (has_error(success, err, timer)) return

      call e%set("nblk", matrix_h%nblk, success)
      if (has_error(success, err, timer)) return

      call e%set("mpi_comm_parent", mpi_comm_parent, success)
      if (has_error(success, err, timer)) return

      call e%set("process_row", matrix_h%prow, success)
      if (has_error(success, err, timer)) return

      call e%set("process_col", matrix_h%pcol, success)
      if (has_error(success, err, timer)) return

      if (debug) then
        call e%set("timings", 1, success)
        if (has_error(success, err, timer)) return
      end if

      success = e%setup()
      if (has_error(success, err, timer)) return

      if (solver == 1) then
        call e%set("solver", ELPA_SOLVER_1STAGE, success)
      else if (solver == 2) then
        call e%set("solver", ELPA_SOLVER_2STAGE, success)
      else
        call err%add_error("Invalid value for ELPA solver")
      end if
      if (has_error(success, err, timer)) return

#:if defined('WITH_ELPA_OPENMP')
      call e%set("omp_threads", omp_get_max_threads(), success)
      if (has_error(success, err, timer)) return
#:endif

      call timer%start("Call ELPA")

      if (compute_vectors) then
        call e%eigenvectors(matrix_h%values, singular_values, matrix_right%values, success)
      else
        call e%eigenvalues(matrix_h%values, singular_values, success)
      end if

      call timer%stop("Call ELPA")

      call matrix_h%dealloc()

      if (success /= ELPA_OK) then
        call err%add_error("ELPA calculation error", success)
        return
      end if

      if (debug .and. rank0) then
        call e%print_times()
      end if

      call elpa_deallocate(e)
      call elpa_uninit()
    end subroutine
#:endfor

    function has_error(success, err, timer)
      integer, intent(in) :: success
      type(error_t), intent(inout) :: err
      type(timer_t), intent(inout) :: timer
      logical :: has_error

      if (success /= ELPA_OK) then
        call err%add_error("ELPA initilization error", success)
        call timer%stop()
      end if
      has_error = err%has_error()
    end function

end module
