#:include 'common.fypp'
module matrix
  use constants
  use debug, only: error_t
  implicit none

  private
  public :: matrix_size
#:for id in IDS
  public :: matrix_${id}$_t
#:endfor

#:for id, dtype, prec in TYPES
  type matrix_${id}$_t
    integer :: rows, cols
    integer :: na_rows, na_cols
    integer :: np_rows, np_cols, prow, pcol
    integer :: nblk
    integer :: desc(desc_size)
    logical :: is_allocated = .false.
    ${dtype}$(${prec}$), pointer :: values(:,:) => null()

    contains
      procedure :: alloc => matrix_alloc_${id}$
      procedure :: alloc_like => matrix_alloc_like_${id}$
      procedure :: alloc_desc => matrix_alloc_desc_${id}$
      procedure :: dealloc => matrix_dealloc_${id}$
      procedure :: get_context => matrix_get_context_${id}$
      procedure :: infog2l => matrix_infog2l_${id}$
      procedure :: indxl2g => matrix_indxl2g_${id}$
      procedure :: set => matrix_set_${id}$
      procedure :: set_at => matrix_set_at_${id}$
      procedure :: set_diag => matrix_set_diag_${id}$
      procedure :: copy_to_array => matrix_copy_to_array_${id}$
      procedure :: copy_from_array => matrix_copy_from_array_${id}$
      procedure :: add => matrix_add_${id}$
      procedure :: sub => matrix_sub_${id}$
      procedure :: copy_from => matrix_copy_from_${id}$
      procedure :: factor => matrix_factor_${id}$
      procedure :: transp => matrix_transp_${id}$
      procedure :: mm => matrix_mm_${id}$
      procedure :: transp_mm => matrix_transp_mm_${id}$
      procedure :: mm_transp => matrix_mm_transp_${id}$
      procedure :: norm => matrix_norm_${id}$
      procedure :: flip_cols => matrix_flip_cols_${id}$
      final :: matrix_finalize_${id}$
  end type
#:endfor

  contains

#:for id, dtype, prec in TYPES
#:set one = CONST(1.0, dtype, prec)
#:set zero = CONST(0.0, dtype, prec)

    !> Allocate a new matrix given the context, the blocking factor 'nblk', and the global size
    subroutine matrix_alloc_${id}$(this, context, nblk, rows, cols, err)
      class(matrix_${id}$_t), intent(inout) :: this
      integer, intent(in) :: context, nblk, rows, cols
      type(error_t), intent(inout) :: err

      integer :: info
      integer, external :: numroc

      this%rows = rows
      this%cols = cols
      this%nblk = nblk

      call blacs_gridinfo(context, this%np_rows, this%np_cols, this%prow, this%pcol)
      this%na_rows = numroc(this%rows, this%nblk, this%prow, 0, this%np_rows)
      this%na_cols = numroc(this%cols, this%nblk, this%pcol, 0, this%np_cols)

      if (this%na_rows == 0 .or. this%na_cols == 0) then
        call err%add_error("Matrix is too small for the given number of processes")
        return
      end if

      call descinit(this%desc, this%rows, this%cols, this%nblk, this%nblk, 0, 0, context, this%na_rows, info)
      if (err%check_info(info, "Error in descinit")) return

      allocate(this%values(this%na_rows, this%na_cols))
      this%is_allocated = .true.
    end subroutine

    !> Allocate a new matrix with the same settings as 'other'
    !> Optionally choose a different size
    subroutine matrix_alloc_like_${id}$(this, other, err, rows, cols)
      class(matrix_${id}$_t), intent(inout) :: this
      class(matrix_${id}$_t), intent(in) :: other
      type(error_t), intent(inout) :: err
      integer, intent(in), optional :: rows, cols

      if (present(rows) .and. present(cols)) then
        call this%alloc(other%desc(desc_context), other%nblk, rows, cols, err)
      else
        call this%alloc(other%desc(desc_context), other%nblk, other%rows, other%cols, err)
      end if
    end subroutine

    !> Allocate a new matrix given the data and a descriptor
    subroutine matrix_alloc_desc_${id}$(this, value_array, desc, err)
      class(matrix_${id}$_t), intent(inout) :: this
      ${dtype}$(${prec}$), intent(in) :: value_array(:,:)
      integer, intent(in) :: desc(desc_size)
      type(error_t), intent(inout) :: err

      integer, external :: numroc

      this%rows = desc(desc_rows)
      this%cols = desc(desc_cols)

      if (desc(desc_nblk_row) /= desc(desc_nblk_col)) then
        call err%add_error("Block sizes in row and column direction must be equal")
        return
      end if

      if (desc(desc_rsrc) /= 0 .or. desc(desc_csrc) /= 0) then
        call err%add_error("Process row and column of the first matrix element must be 0")
        return
      end if

      this%nblk = desc(desc_nblk_row)
      this%desc(:) = desc(:)

      call blacs_gridinfo(desc(desc_context), this%np_rows, this%np_cols, this%prow, this%pcol)
      this%na_rows = numroc(this%rows, this%nblk, this%prow, desc(desc_rsrc), this%np_rows)
      this%na_cols = numroc(this%cols, this%nblk, this%pcol, desc(desc_csrc), this%np_cols)

      if (this%na_rows == 0 .or. this%na_cols == 0) then
        call err%add_error("Matrix is too small for the given number of processes")
        return
      end if

      if (size(value_array, dim=1) /= this%na_rows .or. size(value_array, dim=2) /= this%na_cols) then
        call err%add_error("Local matrix size does not match BLACS layout")
        return
      end if
    end subroutine

    !> Clean-up the matrix
    subroutine matrix_dealloc_${id}$(this)
      class(matrix_${id}$_t), intent(inout) :: this

      if (this%is_allocated) then
        deallocate(this%values)
        this%is_allocated = .false.
      else if (associated(this%values)) then
        nullify(this%values)
      end if
    end subroutine

    !> Finalizer called when allocated matrix_t falls out of scope
    subroutine matrix_finalize_${id}$(this)
      type(matrix_${id}$_t), intent(inout) :: this

      call this%dealloc()
    end subroutine

    !> Returns the BLACS context of the matrix
    function matrix_get_context_${id}$(this) result(context)
      class(matrix_${id}$_t), intent(in) :: this
      integer :: context

      context = this%desc(desc_context)
    end function

    !> Array version of info2gl
    subroutine matrix_infog2l_${id}$(this, local_rows, local_cols, idrows, idcols)
      class(matrix_${id}$_t), intent(in) :: this
      integer, intent(out) :: local_rows(:), local_cols(:)
      integer, intent(out) :: idrows(:), idcols(:)

      call infog2l_helper(this%desc(desc_nblk_row), this%desc(desc_rsrc), this%prow, this%np_rows, local_rows, idrows)
      call infog2l_helper(this%desc(desc_nblk_col), this%desc(desc_csrc), this%pcol, this%np_cols, local_cols, idcols)
    end subroutine

    !> Array version of indxl2g
    subroutine matrix_indxl2g_${id}$(this, global_rows, global_cols)
      class(matrix_${id}$_t), intent(in) :: this
      integer, intent(out) :: global_rows(:), global_cols(:)

      call indxl2g_helper(this%nblk, this%prow, this%np_rows, global_rows)
      call indxl2g_helper(this%nblk, this%pcol, this%np_cols, global_cols)
    end subroutine

    !> Set the diagonal and off-diagonal values
    subroutine matrix_set_${id}$(this, diag, offdiag)
      class(matrix_${id}$_t), intent(inout) :: this
      ${dtype}$(${prec}$), intent(in) :: diag, offdiag

      call p${id}$laset('F', this%rows, this%cols, offdiag, diag, this%values, 1, 1, this%desc)
    end subroutine

    !> Insert a certain value 'val' at a given global position ('global_row','global_col')
    subroutine matrix_set_at_${id}$(this, global_row, global_col, val)
      class(matrix_${id}$_t), intent(inout) :: this
      integer, intent(in) :: global_row, global_col
      ${dtype}$(${prec}$), intent(in) :: val

      integer :: local_row, local_col
      integer :: idrow, idcol

      call infog2l(global_row, global_col, this%desc, this%np_rows, this%np_cols, this%prow, this%pcol,&
          & local_row, local_col, idrow, idcol)

      if (this%prow == idrow .and. this%pcol == idcol) then
        this%values(local_row,local_col) = val
      end if
    end subroutine

    !> Copy the array 'diag' into the diagonal of the matrix
    subroutine matrix_set_diag_${id}$(this, diag)
      class(matrix_${id}$_t), intent(inout) :: this
      ${dtype}$(${prec}$), intent(in) :: diag(:)

      integer :: i, n
      integer, allocatable :: local_rows(:), local_cols(:)
      integer, allocatable :: idrows(:), idcols(:)

      allocate(local_rows(this%rows), idrows(this%rows))
      allocate(local_cols(this%cols), idcols(this%cols))

      call this%infog2l(local_rows, local_cols, idrows, idcols)

      n = min(this%rows, this%cols, size(diag))

      do i = 1, n
        if (this%prow == idrows(i) .and. this%pcol == idcols(i)) then
          this%values(local_rows(i),local_cols(i)) = diag(i)
        end if
      end do
    end subroutine

    !> Copy the distributed values to a local array
    subroutine matrix_copy_to_array_${id}$(this, local)
      class(matrix_${id}$_t), intent(in) :: this
      ${dtype}$(${prec}$), intent(inout) :: local(:,:)

      integer :: global_row, global_col, rows, cols
      integer, allocatable :: local_rows(:), local_cols(:)
      integer, allocatable :: idrows(:), idcols(:)

      allocate(local_rows(this%rows), idrows(this%rows))
      allocate(local_cols(this%cols), idcols(this%cols))

      call this%infog2l(local_rows, local_cols, idrows, idcols)

      rows = size(local, dim=1)
      cols = size(local, dim=2)

      do global_row = 1, rows
        do global_col = 1, cols
          if (this%prow == idrows(global_row) .and. this%pcol == idcols(global_col)) then
            local(global_row,global_col) = this%values(local_rows(global_row),local_cols(global_col))
          else
            local(global_row,global_col) = ${zero}$
          end if
        end do
      end do

      call ${id}$gsum2d(this%desc(desc_context), 'A', ' ', rows, cols, local, rows, -1, -1)
    end subroutine

    !> Copy a local array to the distributed matrix
    subroutine matrix_copy_from_array_${id}$(this, local)
      class(matrix_${id}$_t), intent(inout) :: this
      ${dtype}$(${prec}$), intent(in) :: local(:,:)

      integer :: local_row, local_col
      integer, allocatable :: global_rows(:), global_cols(:)

      allocate(global_rows(this%na_rows), global_cols(this%na_cols))

      call this%indxl2g(global_rows, global_cols)

      do local_row = 1, this%na_rows
        do local_col = 1, this%na_cols
          this%values(local_row,local_col) = local(global_rows(local_row),global_cols(local_col))
        end do
      end do
    end subroutine

    !> Add the matrices 'this' and 'other' and store result in 'this'
    subroutine matrix_add_${id}$(this, other)
      class(matrix_${id}$_t), intent(inout) :: this
      type(matrix_${id}$_t), intent(in) :: other

      call p${id}$geadd('N', other%rows, other%cols, ${one}$, other%values, 1, 1, other%desc, ${one}$,&
          & this%values, 1, 1, this%desc)
    end subroutine

    !> Subtract the matrix 'other' from 'this' and store result in 'this'
    subroutine matrix_sub_${id}$(this, other)
      class(matrix_${id}$_t), intent(inout) :: this
      type(matrix_${id}$_t), intent(in) :: other

      call p${id}$geadd('N', other%rows, other%cols, -${one}$, other%values, 1, 1, other%desc, ${one}$,&
        & this%values, 1, 1, this%desc)
    end subroutine

    !> Copy the matrix 'other' to 'this'
    subroutine matrix_copy_from_${id}$(this, other)
      class(matrix_${id}$_t), intent(inout) :: this
      type(matrix_${id}$_t), intent(in) :: other

      call p${id}$lacpy('F', other%rows, other%cols, other%values, 1, 1, other%desc, this%values, 1, 1, this%desc)
    end subroutine

    !> Multiply the matrix 'this' by a factor = numerator/denominator
    subroutine matrix_factor_${id}$(this, numerator, denominator, err)
      class(matrix_${id}$_t), intent(inout) :: this
      ${dtype}$(${prec}$), intent(in) :: numerator, denominator
      type(error_t), intent(inout) :: err

      integer :: info

      call p${id}$lascl('G', denominator, numerator, this%rows, this%cols, this%values, 1, 1, this%desc, info)
      if (err%check_info(info, "Error in p${id}$lascl")) return
    end subroutine

    !> Transpose (and complex conjugate) the matrix 'other' and store the result in 'this'
    !> Optionally multiply the result by 'factor'
    !> Optionally multiply the result with 'this_factor' instead of overwriting it
    subroutine matrix_transp_${id}$(this, other, factor, this_factor)
      class(matrix_${id}$_t), intent(inout) :: this
      type(matrix_${id}$_t), intent(in) :: other
      ${dtype}$(${prec}$), intent(in), optional :: factor, this_factor

      ${dtype}$(${prec}$) :: fac, this_fac

      fac = ${one}$
      if (present(factor)) then
        fac = factor
      end if

      this_fac = ${zero}$
      if (present(this_factor)) then
        this_fac = this_factor
      end if

#:if dtype == 'real'
      call p${id}$tran(this%rows, this%cols, fac, other%values, 1, 1, other%desc, this_fac, this%values, 1, 1, this%desc)
#:else
      call p${id}$tranc(this%rows, this%cols, fac, other%values, 1, 1, other%desc, this_fac, this%values, 1, 1, this%desc)
#:endif
    end subroutine

    !> Multiply the matrices 'matrix1' and 'matrix2' and store the result in 'this'
    !> Optionally add the result to 'this' instead of overwriting it
    subroutine matrix_mm_${id}$(this, matrix1, matrix2, add)
      class(matrix_${id}$_t), intent(inout) :: this
      type(matrix_${id}$_t), intent(in) :: matrix1, matrix2
      logical, intent(in), optional :: add

      ${dtype}$(${prec}$) :: this_factor

      this_factor = ${zero}$
      if (present(add)) then
        if (add) then
          this_factor = ${one}$
        end if
      end if

      call p${id}$gemm('N', 'N', matrix1%rows, matrix2%cols, matrix1%cols, ${one}$, matrix1%values, 1, 1, matrix1%desc,&
          & matrix2%values, 1, 1, matrix2%desc, this_factor, this%values, 1, 1, this%desc)
    end subroutine

    !> Multiply the matrices transposed_conj('matrix1') and 'matrix2' and store the result in 'this'
    !> Optionally multiply transposed_conj('matrix1') with 'matrix1' instead of 'matrix2'
    !> Optionally add the result to 'this' instead of overwriting it
    subroutine matrix_transp_mm_${id}$(this, matrix1, matrix2, add)
      class(matrix_${id}$_t), intent(inout) :: this
      type(matrix_${id}$_t), intent(in) :: matrix1
      type(matrix_${id}$_t), intent(in), optional :: matrix2
      logical, intent(in), optional :: add

      ${dtype}$(${prec}$) :: this_factor

      this_factor = ${zero}$
      if (present(add)) then
        if (add) then
          this_factor = ${one}$
        end if
      end if

      if (present(matrix2)) then
        call p${id}$gemm('C', 'N', matrix1%cols, matrix2%cols, matrix1%rows, ${one}$, matrix1%values, 1, 1, matrix1%desc,&
            & matrix2%values, 1, 1, matrix2%desc, this_factor, this%values, 1, 1, this%desc)
      else
#:if dtype == 'real'
        call p${id}$syrk('U', 'T', matrix1%cols, matrix1%rows, ${one}$, matrix1%values, 1, 1, matrix1%desc, this_factor,&
            & this%values, 1, 1, this%desc)
#:else
        call p${id}$herk('U', 'C', matrix1%cols, matrix1%rows, ${one}$, matrix1%values, 1, 1, matrix1%desc, this_factor,&
            & this%values, 1, 1, this%desc)
#:endif
      end if
    end subroutine

    !> Multiply the matrices 'matrix1' and transposed_conj('matrix2') and store the result in 'this'
    !> Optionally multiply 'matrix1' with transposed_conj('matrix1') instead of transposed_conj('matrix2')
    !> Optionally add the result to 'this' instead of overwriting it
    subroutine matrix_mm_transp_${id}$(this, matrix1, matrix2, add)
      class(matrix_${id}$_t), intent(inout) :: this
      type(matrix_${id}$_t), intent(in) :: matrix1
      type(matrix_${id}$_t), intent(in), optional :: matrix2
      logical, intent(in), optional :: add

      ${dtype}$(${prec}$) :: this_factor

      this_factor = ${zero}$
      if (present(add)) then
        if (add) then
          this_factor = ${one}$
        end if
      end if

      if (present(matrix2)) then
        call p${id}$gemm('N', 'C', matrix1%rows, matrix2%rows, matrix1%cols, ${one}$, matrix1%values, 1, 1, matrix1%desc,&
            & matrix2%values, 1, 1, matrix2%desc, this_factor, this%values, 1, 1, this%desc)
      else
#:if dtype == 'real'
        call p${id}$syrk('U', 'N', matrix1%rows, matrix1%cols, ${one}$, matrix1%values, 1, 1, matrix1%desc, this_factor,&
            & this%values, 1, 1, this%desc)
#:else
        call p${id}$herk('U', 'N', matrix1%cols, matrix1%rows, ${one}$, matrix1%values, 1, 1, matrix1%desc, this_factor,&
            & this%values, 1, 1, this%desc)
#:endif
      end if
    end subroutine

    !> Return the Frobenius norm of 'this'
    function matrix_norm_${id}$(this)
      class(matrix_${id}$_t), intent(in) :: this
      real(${prec}$) :: matrix_norm_${id}$

      ${dtype}$(${prec}$) :: work(1)
      real(${prec}$), external :: p${id}$lange

      matrix_norm_${id}$ = p${id}$lange('F', this%rows, this%cols, this%values, 1, 1, this%desc, work)
    end function

    !> Invert the order of the columns
    subroutine matrix_flip_cols_${id}$(this, err)
      class(matrix_${id}$_t), intent(inout) :: this
      type(error_t), intent(inout) :: err

      type(matrix_${id}$_t) :: matrix_flipper, matrix_copy
      integer :: i

      call matrix_flipper%alloc_like(this, err, this%cols, this%cols)
      if (err%has_error()) then
        return
      end if

      call matrix_copy%alloc_like(this, err)
      if (err%has_error()) then
        return
      end if

      call matrix_flipper%set(${zero}$, ${zero}$)
      do i = 1, this%cols
        call matrix_flipper%set_at(i, this%cols-i+1, ${one}$)
      end do

      call matrix_copy%copy_from(this)
      call this%mm(matrix_copy, matrix_flipper)

      call matrix_copy%dealloc()
      call matrix_flipper%dealloc()
    end subroutine
#:endfor

    !> Returns the local size of a matrix given its descriptor
    subroutine matrix_size(desc, na_rows, na_cols)
      integer, intent(in) :: desc(desc_size)
      integer, intent(out) :: na_rows, na_cols

      integer :: np_rows, np_cols, prow, pcol
      integer, external :: numroc

      call blacs_gridinfo(desc(desc_context), np_rows, np_cols, prow, pcol)
      na_rows = numroc(desc(desc_rows), desc(desc_nblk_row), prow, desc(desc_rsrc), np_rows)
      na_cols = numroc(desc(desc_cols), desc(desc_nblk_col), pcol, desc(desc_csrc), np_cols)
    end subroutine

    !> Calculates data for infog2l for either rows or columns
    subroutine infog2l_helper(descB, descSRC, proc_position, num_procs, local_indices, local_positions)
      integer, intent(in) :: descB, descSRC
      integer, intent(in) :: proc_position, num_procs
      integer, intent(out) :: local_indices(:), local_positions(:)

      real(dp) :: inv
      integer, dimension(size(local_indices)) :: blk
      integer :: check, i

      inv = 1.0_dp / real(descB, kind=dp)
      do i = 1, size(blk)
        blk(i) = int((i - 1) * inv)
      end do

      local_positions(:) = mod(blk(:) + descSRC, num_procs)
      local_indices(:) = (blk(:) / num_procs + 1) * descB + 1

      check = modulo(proc_position - descSRC, num_procs)
      do i = 1, size(blk)
        if (check >= mod(blk(i), num_procs)) then
          if (proc_position == local_positions(i)) then
            local_indices(i) = local_indices(i) + mod(i - 1, descB)
          end if
          local_indices(i) = local_indices(i) - descB
        end if
      end do
    end subroutine

    !> Calculates the global indices iterating over the local indices
    subroutine indxl2g_helper(nblk, proc_position, num_procs, global_indices)
      integer, intent(in) :: nblk, proc_position, num_procs
      integer, intent(out) :: global_indices(:)

      integer :: i

      do i = 1, size(global_indices)
        global_indices(i) = num_procs * nblk * ((i - 1) / nblk) + mod(i - 1, nblk) +&
            & mod(num_procs + proc_position, num_procs) * nblk + 1
      end do
    end subroutine

end module
