#:include 'common.fypp'
module zolo_svd
  use check
  use constants
  use debug
  use elpa_wrapper
  use ftimings_zsvd, only: timer_t
  use matrix
#:if defined('WITH_MPI_MOD')
  use mpi, only: MPI_COMM_WORLD
#:endif
  use zolo
  implicit none
#:if not defined('WITH_MPI_MOD')
  include 'mpif.h'
#:endif

  private
  public zolo_svd_t
#:for prec in PRECS
  public invert_singular_values_${prec}$
#:endfor

  integer, parameter, public :: ZSVD_R = 1
  integer, parameter, public :: ZSVD_MPI_COMM = 2
  integer, parameter, public :: ZSVD_COMPUTE_LEFT_VECTORS = 3
  integer, parameter, public :: ZSVD_COMPUTE_RIGHT_VECTORS = 4
  integer, parameter, public :: ZSVD_INCREASING_ORDER = 5
  integer, parameter, public :: ZSVD_DECREASING_ORDER = 6
  integer, parameter, public :: ZSVD_COMPUTE_ACCURACY = 7
  integer, parameter, public :: ZSVD_DETECT_INVALID_VALUES = 8
  integer, parameter, public :: ZSVD_PRINT_TIMINGS = 9
  integer, parameter, public :: ZSVD_PRINT_ERRORS = 10
  integer, parameter, public :: ZSVD_ELPA_SOLVER = 11
  integer, parameter, public :: ZSVD_ELPA_DEBUG = 12

  type zolo_svd_t
    private
    real(dp) :: alpha = 0._dp
    real(dp) :: beta = 1._dp
    real(dp) :: accuracy = -1._dp
    integer :: mpi_comm_parent = MPI_COMM_WORLD
    integer :: rk = 1
    logical :: compute_left = .true.
    logical :: compute_right = .true.
    logical :: compute_accuracy = .false.
    logical :: print_timings = .false.
    logical :: print_errors = .true.
    logical :: decreasing_order = .false.
    logical :: detect_invalid_values = .false.
    integer :: elpa_solver = 2
    logical :: elpa_debug = .false.
    integer :: desc_input_matrix(desc_size) = 0
    integer :: desc_left_vectors(desc_size) = 0
    integer :: desc_right_vectors(desc_size) = 0
    type(error_t) :: err

    contains
      procedure :: set_option => zolo_svd_set_option
#:for prec in PRECS
      procedure, private :: zolo_svd_set_bounds_${prec}$
#:endfor
      generic :: set_bounds => zolo_svd_set_bounds_sp, zolo_svd_set_bounds_dp
      procedure :: set_descriptors => zolo_svd_set_descriptors
#:for id in IDS
      procedure, private :: zolo_svd_compute_svd_${id}$
      procedure, private :: zolo_svd_solve_system_${id}$
#:endfor
      generic :: compute_svd => zolo_svd_compute_svd_s, zolo_svd_compute_svd_d, zolo_svd_compute_svd_c, zolo_svd_compute_svd_z
      generic :: solve_system => zolo_svd_solve_system_s, zolo_svd_solve_system_d, zolo_svd_solve_system_c, zolo_svd_solve_system_z
      procedure :: get_accuracy => zolo_svd_get_accuracy
  end type

  abstract interface
#:for prec in PRECS
    !> Interface for the callback function to invert the singular values.
    subroutine invert_singular_values_${prec}$(singular_values)
      import :: ${prec}$
      real(${prec}$), intent(inout) :: singular_values(:)
    end subroutine
#:endfor
  end interface

  contains

    !> Set optional parameters.
    !>   ZSVD_R:                     Order of the Zolotarev functions (Default: 1)
    !>   ZSVD_MPI_COMM:              MPI communicator to be forwarded to ELPA (Default: MPI_COMM_WORLD)
    !>   ZSVD_COMPUTE_LEFT_VECTORS:  Return the matrix of the left singular vectors (Default: true)
    !>   ZSVD_COMPUTE_RIGHT_VECTORS: Return the matrix of the right singular vectors (Default: true)
    !>   ZSVD_INCREASING_ORDER:      Return the singular values in increasing order (Default: true)
    !>   ZSVD_DECREASING_ORDER:      Return the singular values in decreasing order (Default: false)
    !>   ZSVD_COMPUTE_ACCURACY:      Compute the accuracy of the SVD, which can be queries with
    !>                               get_accuracy() (Default: false)
    !>                               Only possible if ZSVD_COMPUTE_LEFT_VECTORS and ZSVD_COMPUTE_RIGHT_VECTORS
    !>   ZSVD_DETECT_INVALID_VALUES: Abort with an error if any singular value is less than zero
    !>   ZSVD_PRINT_TIMINGS:         Print timing information (Default: false)
    !>   ZSVD_PRINT_ERRORS:          Print detailed information about errors (Default: true)
    !>   ZSVD_ELPA_SOLVER:           ELPA solver to be used (Default: 2)
    !>   ZSVD_ELPA_DEBUG:            ELPA debug output (Default: false)
    subroutine zolo_svd_set_option(this, option, val)
      class(zolo_svd_t), intent(inout) :: this
      integer, intent(in) :: option, val

      select case (option)
      case (ZSVD_R)
        this%rk = val
      case (ZSVD_MPI_COMM)
        this%mpi_comm_parent = val
      case (ZSVD_COMPUTE_LEFT_VECTORS)
        this%compute_left = val /= 0
      case (ZSVD_COMPUTE_RIGHT_VECTORS)
        this%compute_right = val /= 0
      case (ZSVD_INCREASING_ORDER)
        this%decreasing_order = val == 0
      case (ZSVD_DECREASING_ORDER)
        this%decreasing_order = val /= 0
      case (ZSVD_COMPUTE_ACCURACY)
        this%compute_accuracy = val /= 0
      case (ZSVD_DETECT_INVALID_VALUES)
        this%detect_invalid_values = val /= 0
      case (ZSVD_PRINT_TIMINGS)
        this%print_timings = val /= 0
      case (ZSVD_PRINT_ERRORS)
        this%print_errors = val /= 0
      case (ZSVD_ELPA_SOLVER)
        this%elpa_solver = val
      case (ZSVD_ELPA_DEBUG)
        this%elpa_debug = val /= 0
      end select
    end subroutine

#:for prec in PRECS
    !> Set the lower and upper bounds of the singular values.
    !> The choice of these parameters influences the solver performance. Better estimates yield less iterations.
    !> Default: lower = 1 and upper = Frobenius norm of the input matrix
    subroutine zolo_svd_set_bounds_${prec}$(this, lower, upper)
      class(zolo_svd_t), intent(inout) :: this
      real(${prec}$), intent(in) :: lower, upper

      this%alpha = real(upper, kind=dp)
      this%beta = real(lower, kind=dp)
    end subroutine
#:endfor

    !> Set the matrix descriptors.
    !> Descriptors of the left and right singular vector matrices are optional,
    !> depending on the choice of ZSVD_COMPUTE_LEFT_VECTORS and ZSVD_COMPUTE_RIGHT_VECTORS
    subroutine zolo_svd_set_descriptors(this, desc_input_matrix, desc_left_vectors, desc_right_vectors)
      class(zolo_svd_t), intent(inout) :: this
      integer, intent(in) :: desc_input_matrix(desc_size)
      integer, intent(in), optional :: desc_left_vectors(desc_size), desc_right_vectors(desc_size)

      this%desc_input_matrix(:) = desc_input_matrix(:)

      if (present(desc_left_vectors)) then
        this%desc_left_vectors(:) = desc_left_vectors(:)
      end if

      if (present(desc_right_vectors)) then
        this%desc_right_vectors(:) = desc_right_vectors(:)
      end if
    end subroutine

#:def ABORT_IF(condition, message)
  if (${condition}$) then
    call this%err%add_error(${message}$)
    info = this%err%get_info()
    if (this%print_errors) then
      call this%err%print()
    end if
    call timer%stop()
    call timer%free()
    return
  end if
#:enddef

#:for id, dtype, prec in TYPES
#:set one = CONST(1.0, dtype, prec)
#:set zero = CONST(0.0, dtype, prec)
    !> Compute the SVD.
    !> Left and right singular vector matrices are optional,
    !> depending on the choice of ZSVD_COMPUTE_LEFT_VECTORS and ZSVD_COMPUTE_RIGHT_VECTORS
    function zolo_svd_compute_svd_${id}$(this, input_matrix, singular_values, left_singular_vectors, right_singular_vectors)&
        & result(info)
      class(zolo_svd_t), intent(inout) :: this
      ${dtype}$(${prec}$), contiguous, target, intent(in) :: input_matrix(:,:)
      real(${prec}$), intent(inout) :: singular_values(:)
      ${dtype}$(${prec}$), contiguous, target, intent(inout), optional :: left_singular_vectors(:,:), right_singular_vectors(:,:)
      integer :: info

      integer :: myid, nprocs
      type(timer_t) :: timer
      type(matrix_${id}$_t) :: matrix_a, matrix_q, matrix_left, matrix_right

      call timer%enable()
      call timer%start("Zolo-SVD")

      call blacs_pinfo(myid, nprocs)

      call this%err%clear()

      @:ABORT_IF(this%rk < 1, "'r' must be greater than zero")

      @:ABORT_IF(this%rk > nprocs, "'r' must not be greater than the number of MPI ranks")

      @:ABORT_IF(all(this%desc_input_matrix == 0), "Array descriptor for input matrix has not been set")

      call matrix_a%alloc_desc(input_matrix, this%desc_input_matrix, this%err)
      @:ABORT_IF(this%err%has_error(), "Error during allocation of matrix A")

      matrix_a%values(1:matrix_a%na_rows, 1:matrix_a%na_cols) => input_matrix

      @:ABORT_IF(matrix_a%cols > matrix_a%rows, "Input matrix must not have more columns than rows")

      if (this%alpha == 0._dp) then
        call timer%start("Estimate upper bound")
        this%alpha = matrix_a%norm()
        call timer%stop("Estimate upper bound")
      end if

      @:ABORT_IF(this%beta <= 0._dp, "The lower bound of the singular values must be greater than zero")

      @:ABORT_IF(this%alpha < this%beta, "Upper bound of the singular values must be greater than the lower bound")

      @:ABORT_IF(size(singular_values) /= matrix_a%cols,"Invalid size of singular values array" )

      if (this%compute_left) then
        @:ABORT_IF(.not. present(left_singular_vectors), "Left singular vector matrix is missing")

        @:ABORT_IF(all(this%desc_left_vectors == 0), "Array descriptor for left singular vector matrix has not been set")

        call matrix_left%alloc_desc(left_singular_vectors, this%desc_left_vectors, this%err)
        @:ABORT_IF(this%err%has_error(), "Error during allocation of left singular vector matrix")

        matrix_left%values(1:matrix_left%na_rows, 1:matrix_left%na_cols) => left_singular_vectors

        @:ABORT_IF(matrix_left%rows /= matrix_a%rows .or. matrix_left%cols /= matrix_a%cols,& 
            & "Invalid shape of left singular vector matrix")

        @:ABORT_IF(matrix_left%nblk /= matrix_a%nblk, "Block size of left singular vector matrix differs from input matrix")
      end if

      ! The right matrix is needed even if only the left is requested
      if (this%compute_left .or. this%compute_right) then
        @:ABORT_IF(.not. present(right_singular_vectors), "Right singular vector matrix is missing")

        @:ABORT_IF(all(this%desc_right_vectors == 0), "Array descriptor for right singular vector matrix has not been set")

        call matrix_right%alloc_desc(right_singular_vectors, this%desc_right_vectors, this%err)
        @:ABORT_IF(this%err%has_error(), "Error during allocation of right singular vector matrix")

        matrix_right%values(1:matrix_right%na_rows, 1:matrix_right%na_cols) => right_singular_vectors

        @:ABORT_IF(matrix_right%rows /= matrix_right%cols .or. matrix_right%cols /= matrix_a%cols,&
            & "Invalid shape of right singular vector matrix")

        @:ABORT_IF(matrix_right%nblk /= matrix_a%nblk, "Block size of right singular vector matrix differs from input matrix")
      end if

      call matrix_q%alloc_like(matrix_a, this%err)
      @:ABORT_IF(this%err%has_error(), "Error during allocation of matrix Q")

      call timer%start("Zolo-PD")
      call zolo_pd_${id}$(matrix_a, matrix_q, this%rk, real(this%alpha, kind=${prec}$), real(this%beta, kind=${prec}$), this%err, timer)
      call timer%stop("Zolo-PD")
      @:ABORT_IF(this%err%has_error(), "Error in Zolo-PD")

      call timer%start("ELPA wrapper")
      call call_elpa_${id}$(matrix_a, matrix_q, singular_values, matrix_right, this%compute_left .or. this%compute_right,&
          & this%elpa_solver, this%mpi_comm_parent, this%err, timer, this%elpa_debug, myid == 0)
      call timer%stop("ELPA wrapper")
      @:ABORT_IF(this%err%has_error(), "Error in ELPA wrapper")

      if (this%compute_left) then
        call timer%start("Compute left matrix")
        call matrix_left%mm(matrix_q, matrix_right)
        call timer%stop("Compute left matrix")
      end if

      if (this%decreasing_order) then
        ! Singular values are returned in increasing order by default
        call timer%start("Reordering")
        singular_values(1:matrix_a%cols) = singular_values(matrix_a%cols:1:-1)

        if (this%compute_left) then
          call matrix_left%flip_cols(this%err)
          @:ABORT_IF(this%err%has_error(), "Error while reordering the left singular vector matrix")
        end if

        if (this%compute_right) then
          call matrix_right%flip_cols(this%err)
          @:ABORT_IF(this%err%has_error(), "Error while reordering the right singular vector matrix")
        end if
        call timer%stop("Reordering")
      end if

      if (this%compute_accuracy) then
        @:ABORT_IF(.not. this%compute_left .or. .not. this%compute_right, "Can only compute accuracy if vectors are present")
        call timer%start("Compute accuracy")
        this%accuracy = svd_accuracy_${id}$(matrix_a, matrix_left, singular_values, matrix_right, this%err)
        call timer%stop("Compute accuracy")
        @:ABORT_IF(this%err%has_error(), "Error during computation of SVD accuracy")
      else
        this%accuracy = -1._dp
      end if

      if (this%detect_invalid_values) then
        @:ABORT_IF(has_invalid_singular_value_${id}$(singular_values, this%err), "Detected at least one invalid singular value")
      end if

      call matrix_a%dealloc()
      call matrix_q%dealloc()
      call matrix_left%dealloc()
      call matrix_right%dealloc()

      call timer%stop("Zolo-SVD")
      if (myid == 0 .and. this%print_timings) then
        call timer%print()
      end if
      call timer%free()
      info = 0
    end function

    !> Solve a system of linear equations using the pseudo inverse:
    !> input_matrix * solution_vector = rhs_vector
    !> Using a callback function to allow user-defined regularization
    function zolo_svd_solve_system_${id}$(this, input_matrix, rhs_vector, solution_vector, invert_singular_values) result(info)
      class(zolo_svd_t), intent(inout) :: this
      ${dtype}$(${prec}$), contiguous, target, intent(in) :: input_matrix(:,:)
      ${dtype}$(${prec}$), intent(in) :: rhs_vector(:)
      procedure(invert_singular_values_${prec}$), optional :: invert_singular_values
      ${dtype}$(${prec}$), intent(out) :: solution_vector(:)

      real(${prec}$), allocatable :: singular_values(:)
      ${dtype}$(${prec}$), allocatable :: solution_vector_aux(:,:)
      integer :: info

      integer :: myid, nprocs, i
      type(timer_t) :: timer
      type(matrix_${id}$_t) :: matrix_a, matrix_q, matrix_left, matrix_right, matrix_singular_values, matrix_pseudoinv
      type(matrix_${id}$_t) :: matrix_vec_rhs, matrix_vec_solution

      call timer%enable()
      call timer%start("Zolo-SVD")

      call blacs_pinfo(myid, nprocs)

      call this%err%clear()

      @:ABORT_IF(this%rk < 1, "'r' must be greater than zero")

      @:ABORT_IF(this%rk > nprocs, "'r' must not be greater than the number of MPI ranks")

      @:ABORT_IF(all(this%desc_input_matrix == 0), "Array descriptor for input matrix has not been set")

      call matrix_a%alloc_desc(input_matrix, this%desc_input_matrix, this%err)
      @:ABORT_IF(this%err%has_error(), "Error during allocation of matrix A")

      matrix_a%values(1:matrix_a%na_rows, 1:matrix_a%na_cols) => input_matrix

      @:ABORT_IF(matrix_a%cols > matrix_a%rows, "Input matrix must not have more columns than rows")

      if (this%alpha == 0._dp) then
        call timer%start("Estimate upper bound")
        this%alpha = matrix_a%norm()
        call timer%stop("Estimate upper bound")
      end if

      @:ABORT_IF(this%beta <= 0._dp, "The lower bound of the singular values must be greater than zero")

      @:ABORT_IF(this%alpha < this%beta, "Upper bound of the singular values must be greater than the lower bound")

      @:ABORT_IF(size(rhs_vector) /= matrix_a%rows, "Length of right-hand side vector must be equal to matrix row count")
      @:ABORT_IF(size(solution_vector) /= matrix_a%cols, "Length of solution vector must be equal to matrix column count")

      allocate(singular_values(matrix_a%cols))

      call matrix_q%alloc_like(matrix_a, this%err)
      @:ABORT_IF(this%err%has_error(), "Error during allocation of matrix Q")

      call matrix_right%alloc_like(matrix_a, this%err, matrix_a%cols, matrix_a%cols)
      @:ABORT_IF(this%err%has_error(), "Error during allocation of right singular vector matrix")

      call timer%start("Zolo-PD")
      call zolo_pd_${id}$(matrix_a, matrix_q, this%rk, real(this%alpha, kind=${prec}$), real(this%beta, kind=${prec}$), this%err, timer)
      call timer%stop("Zolo-PD")
      @:ABORT_IF(this%err%has_error(), "Error in Zolo-PD")

      call timer%start("ELPA wrapper")
      call call_elpa_${id}$(matrix_a, matrix_q, singular_values, matrix_right, .true., this%elpa_solver, this%mpi_comm_parent,&
            & this%err, timer, this%elpa_debug, myid == 0)
      call timer%stop("ELPA wrapper")
      @:ABORT_IF(this%err%has_error(), "Error in ELPA wrapper")

      call matrix_left%alloc_like(matrix_a, this%err)
      @:ABORT_IF(this%err%has_error(), "Error during allocation of left singular vector matrix")

      call timer%start("Compute left matrix")
      call matrix_left%mm(matrix_q, matrix_right)
      call timer%stop("Compute left matrix")

      call matrix_q%dealloc()

      if (this%compute_accuracy) then
        call timer%start("Compute accuracy")
        this%accuracy = svd_accuracy_${id}$(matrix_a, matrix_left, singular_values, matrix_right, this%err)
        call timer%stop("Compute accuracy")
        @:ABORT_IF(this%err%has_error(), "Error during computation of SVD accuracy")
      else
        this%accuracy = -1._dp
      end if

      if (this%detect_invalid_values) then
        @:ABORT_IF(has_invalid_singular_value_${id}$(singular_values, this%err), "Detected at least one invalid singular value")
      end if

      call matrix_singular_values%alloc_like(matrix_right, this%err)
      @:ABORT_IF(this%err%has_error(), "Error during allocation of singular values matrix")

      call matrix_singular_values%set(${zero}$, ${zero}$)
      if (present(invert_singular_values)) then
        call invert_singular_values(singular_values)
      else
        singular_values(:) = 1._${prec}$ / singular_values(:)
      end if
      call matrix_singular_values%set_diag(${one}$ * singular_values(:))

      deallocate(singular_values)

      call matrix_q%alloc_like(matrix_a, this%err, matrix_a%cols, matrix_a%cols)
      @:ABORT_IF(this%err%has_error(), "Error during allocation of auxiliary matrix")

      call matrix_q%mm(matrix_right, matrix_singular_values)

      call matrix_right%dealloc()
      call matrix_singular_values%dealloc()

      call matrix_pseudoinv%alloc_like(matrix_a, this%err, matrix_a%cols, matrix_a%rows)
      @:ABORT_IF(this%err%has_error(), "Error during allocation of pseudo-inverse matrix")

      call matrix_pseudoinv%mm_transp(matrix_q, matrix_left)

      call matrix_q%dealloc()
      call matrix_left%dealloc()

      call matrix_vec_rhs%alloc_like(matrix_a, this%err, matrix_a%rows, matrix_a%cols)
      @:ABORT_IF(this%err%has_error(), "Error during allocation of right-hand side vector")

      call matrix_vec_rhs%set(${zero}$, ${zero}$)

      do i = 1, matrix_vec_rhs%rows
        call matrix_vec_rhs%set_at(i, 1, rhs_vector(i))
      end do

      call matrix_vec_solution%alloc_like(matrix_a, this%err, matrix_a%cols, matrix_a%cols)
      @:ABORT_IF(this%err%has_error(), "Error during allocation of solution vector")

      call matrix_vec_solution%mm(matrix_pseudoinv, matrix_vec_rhs)

      call matrix_pseudoinv%dealloc()
      call matrix_vec_rhs%dealloc()

      allocate(solution_vector_aux(matrix_a%cols,1))
      call matrix_vec_solution%copy_to_array(solution_vector_aux)
      solution_vector(:) = solution_vector_aux(:,1)
      deallocate(solution_vector_aux)

      call matrix_vec_solution%dealloc()
      call matrix_a%dealloc()

      call timer%stop("Zolo-SVD")
      if (myid == 0 .and. this%print_timings) then
        call timer%print()
      end if
      call timer%free()
      info = 0
    end function
#:endfor

    !> Fetch the accuracy of the SVD.
    !> Will return -1 if ZSVD_COMPUTE_ACCURACY has not been set before.
    function zolo_svd_get_accuracy(this) result(acc)
      class(zolo_svd_t), intent(inout) :: this
      real(dp) :: acc

      if (.not. this%compute_accuracy .or. this%err%has_error()) then
        acc = -1._dp
        return
      end if

      acc = this%accuracy
    end function

end module
