#:include 'common.fypp'
module zolo_svd_python_api
  use constants, only: sp, dp, desc_size
  use debug, only: error_t
  use matrix
  use mpi
  use zolo_svd
  use, intrinsic :: iso_c_binding, only: c_int, c_double, c_funptr, c_f_procpointer
  implicit none
  private
  public :: compute_singular_values, compute_singular_vectors, solve_system

  interface compute_singular_values
#:for prec in PRECS
    module procedure compute_singular_values_${prec}$
#:endfor
  end interface

  interface compute_singular_vectors
#:for prec in PRECS
    module procedure compute_singular_vectors_${prec}$
#:endfor
  end interface

  interface solve_system
#:for prec in PRECS
    module procedure solve_system_${prec}$
#:endfor
  end interface

  abstract interface
    function invert_singular_values_python(sv) result(inverted) bind(C)
      import :: c_double
      real(c_double), value, intent(in) :: sv
      real(c_double) :: inverted
    end function
  end interface

  procedure(invert_singular_values_python), pointer :: invert_singular_values

  contains

    subroutine compute_singular_values_python(rows, cols, matrix, print_errors, sv, info) bind(C)
      integer(c_int), value, intent(in) :: rows, cols
      real(c_double), intent(in) :: matrix(rows,cols)
      integer(c_int), value, intent(in) :: print_errors
      real(c_double), intent(out) :: sv(cols)
      integer(c_int), intent(out) :: info

      info = compute_singular_values_dp(matrix, sv, print_errors)
    end subroutine

    subroutine compute_singular_vectors_python(rows, cols, matrix, print_errors, sv, left, right, info) bind(C)
      integer(c_int), value, intent(in) :: rows, cols
      real(c_double), intent(in) :: matrix(rows,cols)
      integer(c_int), value, intent(in) :: print_errors
      real(c_double), intent(out) :: sv(cols), left(rows,cols), right(cols,cols)
      integer(c_int), intent(out) :: info

      info = compute_singular_vectors_dp(matrix, sv, left, right, print_errors)
    end subroutine

    subroutine solve_system_python(rows, cols, matrix, rhs, invert_singular_values_c, print_errors, solution, info) bind(C)
      integer(c_int), value, intent(in) :: rows, cols
      real(c_double), intent(in) :: matrix(rows,cols)
      real(c_double), intent(in) :: rhs(rows)
      type(c_funptr), value, intent(in) :: invert_singular_values_c
      integer(c_int), value, intent(in) :: print_errors
      real(c_double), intent(out) :: solution(cols)
      integer(c_int), intent(out) :: info

      call c_f_procpointer(invert_singular_values_c, invert_singular_values)

      info = solve_system_dp(matrix, rhs, solution, invert_sv, print_errors)

      nullify(invert_singular_values)
    end subroutine

    subroutine invert_sv(sv)
      implicit none
      real(dp), intent(inout) :: sv(:)
      integer :: i

      do i = 1, size(sv)
        sv(i) = invert_singular_values(sv(i))
      end do
    end subroutine

#:for prec in PRECS
    !> Compute the singular values for a given matrix
    function compute_singular_values_${prec}$(array, singular_values, print_errors) result(info)
      real(${prec}$), intent(in) :: array(:,:)
      real(${prec}$), intent(out) :: singular_values(:)
      integer, intent(in) :: print_errors
      integer :: info

      logical :: initialized
      integer :: ierr, myid, nprocs, rows, cols
      integer :: np_rows, np_cols, my_prow, my_pcol
      integer :: context, nblk
      integer, parameter :: nblk_default = 64
      type(zolo_svd_t) :: svd
#:if prec == 'sp'
      type(matrix_s_t) :: matrix_a
#:else
      type(matrix_d_t) :: matrix_a
#:endif
      type(error_t) :: err

      integer, external :: numroc

      call mpi_initialized(initialized, ierr)
      if (ierr /= MPI_SUCCESS) then
        info = 1000
        return
      end if
      if (.not. initialized) then
        info = 1000
        return
      end if

      call blacs_pinfo(myid, nprocs)

      do np_cols = nint(sqrt(real(nprocs))), 2, -1
        if (mod(nprocs, np_cols) == 0) exit
      end do
      np_rows = nprocs / np_cols

      call blacs_get(-1, 0, context)
      call blacs_gridinit(context, 'R', np_rows, np_cols)
      call blacs_gridinfo(context, np_rows, np_cols, my_prow, my_pcol)

      rows = size(array, dim=1)
      cols = size(array, dim=2)

      nblk = nblk_default
      do while ((nblk > rows .or. nblk > cols) .and. nblk > 2)
        nblk = nblk / 2
      end do

      call matrix_a%alloc(context, nblk, rows, cols, err)
      if (err%has_error()) then
        info = 1
        if (print_errors == 1) then
          call err%print()
        end if
        call blacs_gridexit(context)
        return
      end if

      call matrix_a%copy_from_array(array)

      call svd%set_option(ZSVD_COMPUTE_LEFT_VECTORS, 0)
      call svd%set_option(ZSVD_COMPUTE_RIGHT_VECTORS, 0)
      call svd%set_option(ZSVD_PRINT_ERRORS, print_errors)

      call svd%set_descriptors(matrix_a%desc)

      info = svd%compute_svd(matrix_a%values, singular_values)

      call matrix_a%dealloc()

      call blacs_gridexit(context)
    end function

    !> Compute the singular values and vectors for a given matrix
    function compute_singular_vectors_${prec}$(array, singular_values, left_array, right_array, print_errors) result(info)
      real(${prec}$), intent(in) :: array(:,:)
      real(${prec}$), intent(out) :: singular_values(:), left_array(:,:), right_array(:,:)
      integer, intent(in) :: print_errors
      integer :: info

      logical :: initialized
      integer :: ierr, myid, nprocs, rows, cols
      integer :: np_rows, np_cols, my_prow, my_pcol
      integer :: context, nblk
      integer, parameter :: nblk_default = 64
      type(zolo_svd_t) :: svd
#:if prec == 'sp'
      type(matrix_s_t) :: matrix_a, matrix_left, matrix_right
#:else
      type(matrix_d_t) :: matrix_a, matrix_left, matrix_right
#:endif
      type(error_t) :: err

      integer, external :: numroc

      call mpi_initialized(initialized, ierr)
      if (ierr /= MPI_SUCCESS) then
        info = 1000
        return
      end if
      if (.not. initialized) then
        info = 1000
        return
      end if

      call blacs_pinfo(myid, nprocs)

      do np_cols = nint(sqrt(real(nprocs))), 2, -1
        if (mod(nprocs, np_cols) == 0) exit
      end do
      np_rows = nprocs / np_cols

      call blacs_get(-1, 0, context)
      call blacs_gridinit(context, 'R', np_rows, np_cols)
      call blacs_gridinfo(context, np_rows, np_cols, my_prow, my_pcol)

      rows = size(array, dim=1)
      cols = size(array, dim=2)

      nblk = nblk_default
      do while ((nblk > rows .or. nblk > cols) .and. nblk > 2)
        nblk = nblk / 2
      end do

      call matrix_a%alloc(context, nblk, rows, cols, err)
      if (err%has_error()) then
        info = 1
        if (print_errors == 1) then
          call err%print()
        end if
        call blacs_gridexit(context)
        return
      end if

      call matrix_left%alloc_like(matrix_a, err)
      if (err%has_error()) then
        info = 1
        if (print_errors == 1) then
          call err%print()
        end if
        call blacs_gridexit(context)
        return
      end if

      call matrix_right%alloc_like(matrix_a, err, matrix_a%cols, matrix_a%cols)
      if (err%has_error()) then
        info = 1
        if (print_errors == 1) then
          call err%print()
        end if
        call blacs_gridexit(context)
        return
      end if

      call matrix_a%copy_from_array(array)

      call svd%set_option(ZSVD_PRINT_ERRORS, print_errors)

      call svd%set_descriptors(matrix_a%desc, matrix_left%desc, matrix_right%desc)

      info = svd%compute_svd(matrix_a%values, singular_values, matrix_left%values, matrix_right%values)

      call matrix_a%dealloc()

      call matrix_left%copy_to_array(left_array)
      call matrix_right%copy_to_array(right_array)

      call matrix_left%dealloc()
      call matrix_right%dealloc()

      call blacs_gridexit(context)
    end function

    !> Solve the system of equations given a matrix and a right-hand side vector
    function solve_system_${prec}$(array, rhs, solution, invert_sv_callback, print_errors) result(info)
      real(${prec}$), intent(in) :: array(:,:)
      real(${prec}$), intent(in) :: rhs(:)
      real(${prec}$), intent(out) :: solution(:)
      procedure(invert_singular_values_${prec}$) :: invert_sv_callback
      integer, intent(in) :: print_errors
      integer :: info

      logical :: initialized
      integer :: ierr, myid, nprocs, rows, cols
      integer :: np_rows, np_cols, my_prow, my_pcol
      integer :: context, nblk
      integer, parameter :: nblk_default = 64
      type(zolo_svd_t) :: svd
#:if prec == 'sp'
      type(matrix_s_t) :: matrix_a
#:else
      type(matrix_d_t) :: matrix_a
#:endif
      type(error_t) :: err

      integer, external :: numroc

      call mpi_initialized(initialized, ierr)
      if (ierr /= MPI_SUCCESS) then
        info = 1000
        return
      end if
      if (.not. initialized) then
        info = 1000
        return
      end if

      call blacs_pinfo(myid, nprocs)

      do np_cols = nint(sqrt(real(nprocs))), 2, -1
        if (mod(nprocs, np_cols) == 0) exit
      end do
      np_rows = nprocs / np_cols

      call blacs_get(-1, 0, context)
      call blacs_gridinit(context, 'R', np_rows, np_cols)
      call blacs_gridinfo(context, np_rows, np_cols, my_prow, my_pcol)

      rows = size(array, dim=1)
      cols = size(array, dim=2)

      nblk = nblk_default
      do while ((nblk > rows .or. nblk > cols) .and. nblk > 2)
        nblk = nblk / 2
      end do

      call matrix_a%alloc(context, nblk, rows, cols, err)
      if (err%has_error()) then
        info = 1
        if (print_errors == 1) then
          call err%print()
        end if
        call blacs_gridexit(context)
        return
      end if

      call matrix_a%copy_from_array(array)

      call svd%set_option(ZSVD_PRINT_ERRORS, print_errors)

      call svd%set_descriptors(matrix_a%desc)

      info = svd%solve_system(matrix_a%values, rhs, solution, invert_sv_callback)

      call matrix_a%dealloc()

      call blacs_gridexit(context)
    end function
#:endfor

end module
