import sys, os
sys.setdlopenflags(os.RTLD_NOW | os.RTLD_GLOBAL)
import numpy as np
import zolo_svd
from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.rank

def create_matrix(rows, cols):
    matrix = None
    if rank == 0:
        np.random.seed(42)
        matrix = np.random.rand(rows, cols)
    return comm.bcast(matrix, root=0)

def numpy_svd(matrix):
    u, sv, vh = np.linalg.svd(matrix, full_matrices=False)
    return sv[::-1], u[:,::-1], vh.transpose()[:,::-1]

def check_singular_values(matrix):
    sv = zolo_svd.singular_values(matrix)

    if rank == 0:
        sv_ref, _, _ = numpy_svd(matrix)
        err = np.max(np.abs(sv - sv_ref) / sv_ref)
        if err > 1.e-12:
            raise Exception("Error too large: {}".format(err))

def check_singular_vectors(matrix):
    sv, u, v = zolo_svd.singular_vectors(matrix)

    if rank == 0:
        sv_ref, u_ref, v_ref = numpy_svd(matrix)

        for i in range(matrix.shape[1]):
            x = v[:,i]
            if x[0] * v_ref[0,i] < 0:
                x = -x
            err = np.max(np.abs((x - v_ref[:,i]) / v_ref[:,i]))
            if err > 1.e-8:
                raise Exception("Error too large: {} {}".format(i, err))

            x = u[:,i]
            if x[0] * u_ref[0,i] < 0:
                x = -x
            err = np.max(np.abs((x - u_ref[:,i]) / u_ref[:,i]))
            if err > 1.e-8:
                raise Exception("Error too large: {} {}".format(i, err))

def check_solve(matrix):
    x_ref = np.random.rand(matrix.shape[1])
    x_ref = comm.bcast(x_ref, root=0)

    def invert_singular_values(sv):
        return 1 / sv

    b = np.matmul(matrix, x_ref)
    x = zolo_svd.solve(matrix, b, invert_singular_values)

    if rank == 0:
        err = np.max(np.abs(x - x_ref) / x_ref)
        if err > 1.e-9:
            raise Exception("Error too large: {}".format(err))


for rows,cols in [(77,55), (55,55)]:
    matrix = create_matrix(rows, cols)
    check_singular_values(matrix)
    check_singular_vectors(matrix)
    check_solve(matrix)

try:
    matrix = np.zeros((55,77))
    sv = zolo_svd.singular_values(matrix)
except ValueError:
    pass
else:
    raise Exception("No error thrown")
