#:include '../src/common.fypp'
program test_zolo_svd_scalapack_api
  use constants, only: sp, dp, desc_size, pi
#:if defined('WITH_MPI_MOD')
  use mpi
#:endif
  use zolo_svd_scalapack_api
  implicit none
#:if not defined('WITH_MPI_MOD')
  include 'mpif.h'
#:endif

  integer :: ierr, myid, nprocs

  call mpi_init(ierr)
  call mpi_comm_rank(MPI_COMM_WORLD, myid, ierr)
  call mpi_comm_size(MPI_COMM_WORLD, nprocs, ierr)

#:for id in IDS
  call run_tests_${id}$(myid, nprocs, 7, 4, 'R')
  call run_tests_${id}$(myid, nprocs, 200, 16, 'R')
  call run_tests_${id}$(myid, nprocs, 200, 16, 'C')
#:endfor

  if (myid == 0) then
    write(*,'(a)') "SUCCESS"
  end if

  call mpi_finalize(ierr)

  contains

#:for id, dtype, prec in TYPES
    !> Call SVD from ScaLAPACK
    subroutine scalapack_${id}$(myid, rows, cols, array_input, desc, singular_values, array_left, array_right)
      integer, intent(in) :: myid, rows, cols
      ${dtype}$(${prec}$), intent(in) :: array_input(:,:)
      integer, intent(in) :: desc(desc_size)
      real(${prec}$), intent(inout) :: singular_values(:)
      ${dtype}$(${prec}$), intent(inout) :: array_left(:,:), array_right(:,:)

      integer :: na_rows, na_cols, info, lwork
      ${dtype}$(${prec}$), allocatable :: array(:,:), work(:)
      ${dtype}$(${prec}$) :: rtmp(1)
#:if dtype == 'complex'
      real(${prec}$), allocatable :: rwork(:)
      real(${prec}$) :: rrtmp(1)
#:endif
      real(dp) :: time

      na_rows = size(array_input, dim=1)
      na_cols = size(array_input, dim=2)

      allocate(array(na_rows,na_cols))
      array(:,:) = array_input(:,:)

      call p${id}$gesvd('V', 'V', rows, cols, array, 1, 1, desc, singular_values, array_left, 1, 1, desc, array_right, 1, 1, desc,&
          & rtmp, -1,&
#:if dtype == 'complex'
          & rrtmp,&
#:endif
          & info)
      if (info /= 0) then
        write(*,'(a)') "Error in p${id}$gesvd"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      lwork = int(rtmp(1)) + 1
      allocate(work(lwork))
#:if dtype == 'complex'
      allocate(rwork(int(rrtmp(1)) + 1))
#:endif

      time = mpi_wtime()
      call p${id}$gesvd('V', 'V', rows, cols, array, 1, 1, desc, singular_values, array_left, 1, 1, desc, array_right, 1, 1, desc,&
          & work, lwork,&
#:if dtype == 'complex'
          & rwork,&
#:endif
          & info)
      time = mpi_wtime() - time

      if (info /= 0) then
        write(*,'(a)') "Error in p${id}$gesvd"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      if (myid == 0) then
        write(*,'(a,1x,f012.9,1x,a)') "Time for ScaLAPACK:", time, "s"
      end if
    end subroutine

    subroutine run_tests_${id}$(myid, nprocs, rows, nblk, blacs_layout)
      integer, intent(in) :: myid, nprocs, rows, nblk
      character, intent(in) :: blacs_layout

      integer :: info, cols, i, j
      integer :: np_cols, np_rows, my_prow, my_pcol, context, na_rows, na_cols
      integer :: local_row, local_col, idrow, idcol
      integer :: desc(desc_size)
      ${dtype}$(${prec}$) :: val
      ${dtype}$(${prec}$), allocatable :: array(:,:), array_left(:,:), array_right(:,:)
      real(${prec}$), allocatable :: singular_values(:), singular_values_ref(:)

      integer, external :: numroc

      cols = rows

      do np_cols = nint(sqrt(real(nprocs))), 2, -1
        if (mod(nprocs, np_cols) == 0) exit
      end do
      np_rows = nprocs / np_cols

      if (myid == 0) then
        write(*,'(a)') "Type=${dtype}$, Precision=${prec}$"
        write(*,'(2(a,i0))') "Matrix size=", rows, "x", cols
        write(*,'(3(a,i0))') "Number of processors: rows=", np_rows, ", cols=", np_cols, ", total=", nprocs
      end if

      call blacs_get(-1, 0, context)
      call blacs_gridinit(context, blacs_layout, np_rows, np_cols)
      call blacs_gridinfo(context, np_rows, np_cols, my_prow, my_pcol)

      na_rows = numroc(rows, nblk, my_prow, 0, np_rows)
      na_cols = numroc(cols, nblk, my_pcol, 0, np_cols)

      if (na_rows == 0 .or. na_cols == 0) then
        write(*,'(a)') "Test matrix is too small for the given number of processes"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      call descinit(desc, rows, cols, nblk, nblk, 0, 0, context, na_rows, info)
      if (info /= 0) then
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      allocate(array(na_rows,na_cols))
      allocate(array_left(na_rows,na_cols), array_right(na_rows,na_cols))

      do i = 1, rows
        do j = 1, cols
#:if dtype == 'real'
          val = real(9.0 / (1.0 + (i-j)**2), kind=${prec}$)
#:else
          val = cmplx(9.0 / (1.0 + (i-j)**2), 3.0 / (1.0 + (i-j)**2), kind=${prec}$)
#:endif
          call infog2l(i, j, desc, np_rows, np_cols, my_prow, my_pcol, local_row, local_col, idrow, idcol)
          if (my_prow == idrow .and. my_pcol == idcol) then
            array(local_row,local_col) = val
          end if
        end do
      end do

      allocate(singular_values(cols), singular_values_ref(cols))

      call scalapack_${id}$(myid, rows, cols, array, desc, singular_values_ref, array_left, array_right)

      call test_with_vectors_${id}$(myid, rows, cols, array, desc, singular_values, array_left, array_right)
      call check_result_${id}$(myid, singular_values, singular_values_ref)

      call test_only_left_vectors_${id}$(myid, rows, cols, array, desc, singular_values, array_left, array_right)
      call check_result_${id}$(myid, singular_values, singular_values_ref)

      call test_only_right_vectors_${id}$(myid, rows, cols, array, desc, singular_values, array_right)
      call check_result_${id}$(myid, singular_values, singular_values_ref)

      call test_without_vectors_${id}$(myid, rows, cols, array, desc, singular_values)
      call check_result_${id}$(myid, singular_values, singular_values_ref)

      call test_invalid_${id}$(rows, cols)

      call blacs_gridexit(context)
    end subroutine

    subroutine test_with_vectors_${id}$(myid, rows, cols, array_input, desc, singular_values, array_left, array_right,&
          & request_right)
      integer, intent(in) :: myid, rows, cols
      ${dtype}$(${prec}$), intent(in) :: array_input(:,:)
      integer, intent(in) :: desc(desc_size)
      real(${prec}$), intent(inout) :: singular_values(:)
      ${dtype}$(${prec}$), intent(inout) :: array_left(:,:), array_right(:,:)
      character, intent(in), optional :: request_right

      integer :: na_rows, na_cols, info
      ${dtype}$(${prec}$), allocatable :: array(:,:)
      ${dtype}$(${prec}$) :: work(1)
#:if dtype == 'complex'
      real(${prec}$) :: rwork(1)
#:endif
      real(dp) :: time
      character :: right

      na_rows = size(array_input, dim=1)
      na_cols = size(array_input, dim=2)

      allocate(array(na_rows,na_cols))
      array(:,:) = array_input(:,:)

      right = 'V'
      if (present(request_right)) then
        right = request_right
      end if

      time = mpi_wtime()
      call p${id}$gesvd_zolo('V', right, rows, cols, array, 1, 1, desc, singular_values, array_left, 1, 1, desc, array_right, 1, 1,&
          & desc, work, 1,&
#:if dtype == 'complex'
          & rwork,&
#:endif
          & info)
      time = mpi_wtime() - time

      if (info /= 0) then
        write(*,'(a)') "Error in p${id}$gesvd_zolo"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      if (right == 'V') then
        if (myid == 0) then
          write(*,'(a,1x,g013.6e2)') "Accuracy of SVD:", real(work(1))
        end if
#:if prec == 'sp'
        if (real(work(1), kind=dp) > 1.e-4_dp) then
#:else
        if (real(work(1), kind=dp) > 1.e-12_dp) then
#:endif
          call mpi_abort(MPI_COMM_WORLD, 1, info)
        end if
      end if

      if (myid == 0) then
        write(*,'(a,1x,f012.9,1x,a)') "Time for Zolo-SVD:", time, "s"
      end if
    end subroutine

    subroutine test_only_left_vectors_${id}$(myid, rows, cols, array_input, desc, singular_values, array_left, array_right)
      integer, intent(in) :: myid, rows, cols
      ${dtype}$(${prec}$), intent(in) :: array_input(:,:)
      integer, intent(in) :: desc(desc_size)
      real(${prec}$), intent(inout) :: singular_values(:)
      ${dtype}$(${prec}$), intent(inout) :: array_left(:,:), array_right(:,:)

      call test_with_vectors_${id}$(myid, rows, cols, array_input, desc, singular_values, array_left, array_right, 'N')
    end subroutine

    subroutine test_only_right_vectors_${id}$(myid, rows, cols, array_input, desc, singular_values, array_right)
      integer, intent(in) :: myid, rows, cols
      ${dtype}$(${prec}$), intent(in) :: array_input(:,:)
      integer, intent(in) :: desc(desc_size)
      real(${prec}$), intent(inout) :: singular_values(:)
      ${dtype}$(${prec}$), intent(inout) :: array_right(:,:)

      integer :: na_rows, na_cols, info
      ${dtype}$(${prec}$), allocatable :: array(:,:)
      ${dtype}$(${prec}$) :: work(1), nope(1)
#:if dtype == 'complex'
      real(${prec}$) :: rwork(1)
#:endif
      real(dp) :: time

      na_rows = size(array_input, dim=1)
      na_cols = size(array_input, dim=2)

      allocate(array(na_rows,na_cols))
      array(:,:) = array_input(:,:)

      time = mpi_wtime()
      call p${id}$gesvd_zolo('N', 'V', rows, cols, array, 1, 1, desc, singular_values, nope, 1, 1, desc, array_right, 1, 1, desc,&
          & work, 1,&
#:if dtype == 'complex'
          & rwork,&
#:endif
          & info)
      time = mpi_wtime() - time

      if (info /= 0) then
        write(*,'(a)') "Error in p${id}$gesvd_zolo"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      if (myid == 0) then
        write(*,'(a,1x,f012.10,1x,a)') "Time for Zolo-SVD:", time, "s"
      end if
    end subroutine

    subroutine test_without_vectors_${id}$(myid, rows, cols, array_input, desc, singular_values)
      integer, intent(in) :: myid, rows, cols
      ${dtype}$(${prec}$), intent(in) :: array_input(:,:)
      integer, intent(in) :: desc(desc_size)
      real(${prec}$), intent(inout) :: singular_values(:)

      integer :: na_rows, na_cols, info
      ${dtype}$(${prec}$), allocatable :: array(:,:)
      ${dtype}$(${prec}$) :: work(1), nope(1)
#:if dtype == 'complex'
      real(${prec}$) :: rwork(1)
#:endif
      real(dp) :: time

      na_rows = size(array_input, dim=1)
      na_cols = size(array_input, dim=2)

      allocate(array(na_rows,na_cols))
      array(:,:) = array_input(:,:)

      time = mpi_wtime()
      call p${id}$gesvd_zolo('N', 'N', rows, cols, array, 1, 1, desc, singular_values, nope, 1, 1, desc, nope, 1, 1, desc, work, 1,&
#:if dtype == 'complex'
          & rwork,&
#:endif
          & info)
      time = mpi_wtime() - time

      if (info /= 0) then
        write(*,'(a)') "Error in p${id}$gesvd_zolo"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      if (myid == 0) then
        write(*,'(a,1x,f012.10,1x,a)') "Time for Zolo-SVD:", time, "s"
      end if
    end subroutine

    subroutine test_invalid_${id}$(rows, cols)
      integer, intent(in) :: rows, cols

      integer :: info, desc(1)
      ${dtype}$(${prec}$) :: nope(1)
      real(${prec}$) :: missing(1)
#:if dtype == 'complex'
      real(${prec}$) :: rwork(1)
#:endif

      call p${id}$gesvd_zolo('V', 'V', rows, 2*cols, nope, 1, 1, desc, missing, nope, 1, 1, desc, nope, 1, 1, desc, nope, 1,&
#:if dtype == 'complex'
          & rwork,&
#:endif
          & info)

      if (info == 0) then
        write(*,'(a)') "No error in p${id}$gesvd_zolo, although an error was expected"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if
    end subroutine

    subroutine check_result_${id}$(myid, singular_values, singular_values_ref)
      integer, intent(in) :: myid
      real(${prec}$), intent(in) :: singular_values(:), singular_values_ref(:)

      integer :: info
      real(${prec}$) :: val

      if (myid == 0) then
        val = maxval(abs(singular_values - singular_values_ref) / singular_values_ref)
        write(*,'(a,1x,g013.6e2)') "Largest relative deviation between ScaLAPACK and Zolo-SVD:", val
#:if prec == 'sp'
        if (val > 1.e-4_dp) then
#:else
        if (val > 1.e-12_dp) then
#:endif
          call mpi_abort(MPI_COMM_WORLD, 1, info)
        end if
      end if
    end subroutine
#:endfor

end program
